#include <sys/time.h>
#include <stdlib.h>

#include "common.c"

void init_seed(void)
{
    struct timeval tv;
    struct timezone tz;
    unsigned int seed = 0;

    if (gettimeofday(&tv, &tz) < 0) {
	perror("gettimeofday");
	seed = 1;
    } else {
	seed = (unsigned int)(tv.tv_usec % (long)RAND_MAX);
    }
    srandom(seed);
}

uint16_t decide_port(uint16_t dst)
{
    uint16_t port = 0;
    int sock = 0;
    socklen_t sin6_len = 0;
    struct sockaddr_in6 sin6;

    sock = socket(PF_INET6, SOCK_DGRAM, IPPROTO_UDP);
    if (sock < 0) {
	printf("cannot open INET6 socket.\n");
	return 0;
    }
    sin6_len = sizeof(sin6);
    memset(&sin6, 0, sin6_len);
    sin6.sin6_family = AF_INET6;
    while (port == 0) {
	port = (uint16_t)(random() % (long)65535);
	if ((port <= 1024) || (port == dst)) {
		port = 0;
		continue;
	}
	sin6.sin6_port = htons(port);
 	if (bind(sock, (struct sockaddr*)&sin6, sin6_len) == -1) {
		printf("port = %d, ", port);
        	perror("decide_port");
		port = 0;
		close(sock);
		continue;
	}
    }
    close(sock);
    return port;
}


int main(int ac,char **av)
{
    __u32 flowlabel = 0;
    int rsock, ssock, len, err_num, itmp;
    socklen_t sin6_len = 0;
    uint16_t port, d_port;
    struct in6_addr addr;
    struct sockaddr_in6 sin6;
    char *target, v6addr[128];

    rsock = ssock = len = err_num = itmp = 0;
    port = d_port = 0;
/*
    if ((ac != 5) && (ac != 6)) {
	printf("usage: %s <flowlabel> <dst_host> <dst_port> <wait_port> [<wait_host>]\n",av[0]);
	return 0;
    }
*/
    if ((ac != 4) && (ac != 5)) {
	printf("usage: %s <flowlabel> <dst_host> <dst_port> [<wait_host>]\n",av[0]);
	return 0;
    }

    sscanf(av[1],"%x",&flowlabel);
    if (flowlabel == 0) {
	printf("flowlabel is null.\n");
	return -1;
    }
    target = av[2];
    sscanf(av[3],"%d",&itmp);
    if (itmp == 0) {
	printf("dst_port is null.\n");
	return -2;
    } else {
	d_port = (uint16_t)itmp;
	itmp = 0;
    }
/*
    sscanf(av[4],"%d",&itmp);
    if (itmp == 0) {
	printf("wait_port is null.\n");
	return -2;
    } else {
	port = (uint16_t)itmp;
    }
*/
    init_seed();
    port = decide_port(d_port);

    if (getip6name(&addr, target, "dst_addr") < 0) return -3;

    sin6_len = sizeof(sin6);
    memset(&sin6, 0, sin6_len);
    sin6.sin6_family = AF_INET6;
    sin6.sin6_port = htons(d_port);
    memcpy(&sin6.sin6_addr, &addr, 16);
    rsock = socket(PF_INET6, SOCK_DGRAM, IPPROTO_UDP);
    ssock = socket(PF_INET6, SOCK_DGRAM, IPPROTO_UDP);
    if ((rsock < 0) || (ssock < 0)) {
	printf("cannot open INET6 socket.\n");
	return -4;
    }
    if (set_flowlabel(ssock, &sin6, flowlabel) < 0) {
	printf("cannot use flowlabel.\n");
	close(rsock);
	close(ssock);
	return -5;
    }
    memset(v6addr, 0, 128);
    sprintf(v6addr, "%d", port);
    len = sendto(ssock,v6addr,strlen(v6addr),0,(struct sockaddr*)&sin6,sin6_len);
    printf("sending...\n");
    if (len < 0) {
	perror("send");
	err_num = -7;
	goto err;
    }
    release_flowlabel(ssock, &sin6, flowlabel);
    close(ssock); ssock = 0;

    if (ac == 5) {
//    if (ac == 6) {
//	if (getip6name(&addr, av[5], "src_addr") < 0) return -3;
	if (getip6name(&addr, av[4], "src_addr") < 0) return -3;
	memcpy(&sin6.sin6_addr, &addr, 16);
    } else {
        if (getsockname(rsock, (struct sockaddr *)&sin6, &sin6_len) < 0) {
	    perror("getsockname");
	    return -20;
        }
        memset(v6addr, 0, 128);
        if (inet_ntop(AF_INET6, &sin6.sin6_addr, v6addr, 128) == NULL) {
	    perror("inet_ntop");
	    memcpy(v6addr, target, (strlen(target) < 127 ? strlen(target) : 127));
        }
        printf("src_addr = %s\n",v6addr);
    }
    sin6.sin6_port = htons(port);
    if (set_rxflow(rsock) < 0) {
	printf("can't get flowlabel\n");
    }
    if (bind(rsock, (struct sockaddr*)&sin6, sin6_len) == -1) {
        perror("bind");
        err_num = -10;
        goto err;
    }
    len = 0; memset(v6addr, 0, 128);
/* cannot get flowinfo
    len = recvfrom(rsock, v6addr, 128, 0, (struct sockaddr*)&sin6, &sin6_len);
*/
    len = recvlabel(rsock, v6addr, 128, 0, (struct sockaddr*)&sin6, &sin6_len);
    if (len < 0) {
	perror("recvfrom");
	err_num = -8;
	goto err;
    }
    printf("dst_host flow has 0x%x.\n",ntohl(sin6.sin6_flowinfo));
    printf("message: %s\n", v6addr);
 err:
    if (ssock != 0) {
	release_flowlabel(ssock, &sin6, flowlabel);
	close(ssock);
    }
    close(rsock);
    return err_num;
}
